"""Loader for creating instances of the devices from a config file."""

import asyncio
import inspect
import logging
import os
import time
from pathlib import Path
from typing import Mapping, Sequence

import tomlkit
from ophyd import Device as ThreadedDevice
from ophyd.sim import make_fake_device
from ophyd_async.core import DEFAULT_TIMEOUT, NotConnected
from ophydregistry import Registry

from .devices.aerotech import AerotechStage
from .devices.aps import ApsMachine
from .devices.area_detector import make_area_detector
from .devices.beamline_manager import BeamlineManager
from .devices.dxp import make_dxp_device
from .devices.energy_positioner import EnergyPositioner
from .devices.heater import CapillaryHeater
from .devices.ion_chamber import IonChamber
from .devices.mirrors import HighHeatLoadMirror, KBMirrors
from .devices.motor import Motor, load_motors
from .devices.power_supply import NHQ203MChannel
from .devices.robot import Robot
from .devices.scaler import MultiChannelScaler
from .devices.shutter import PssShutter
from .devices.slits import ApertureSlits, BladeSlits
from .devices.stage import XYStage
from .devices.table import Table
from .devices.xia_pfcu import PFCUFilterBank
from .devices.xspress import make_xspress_device
from .exceptions import InvalidConfiguration

log = logging.getLogger(__name__)


instrument = None


class Instrument:
    """A beamline instrument built from config files of Ophyd devices.

    *device_classes* should be dictionary that maps configuration
     section names to device classes (or similar items).

    Example:

    ```python
    instrument = Instrument({
      "ion_chamber": IonChamber
      "motors": load_motors,
    })
    ```

    The values in *device_classes* should be one of the following:

    1. A device class
    2. A callable that returns an instantiated device object
    3. A callable that returns a sequence of device objects

    Parameters
    ==========
    device_classes
      Maps config section names to device classes.

    """

    devices: list
    registry: Registry
    beamline_name: str = ""
    hardware_is_present: bool | None = None

    def __init__(self, device_classes: Mapping, registry: Registry | None = None):
        self.devices = []
        if registry is None:
            registry = Registry(auto_register=False, use_typhos=False)
        self.registry = registry
        self.device_classes = device_classes

    def parse_toml_file(self, fd):
        """Parse TOML instrument configuration and create devices.

        An open file descriptor
        """
        config = tomlkit.load(fd)
        # Set global parameters
        beamline = config.get("beamline", {})
        self.beamline_name = beamline.get("name", self.beamline_name)
        self.hardware_is_present = beamline.get(
            "hardware_is_present", self.hardware_is_present
        )
        # Make devices from config file
        return self.parse_config(config)

    def parse_config(self, cfg):
        devices = []
        for key, Klass in self.device_classes.items():
            # Create the devices
            for params in cfg.get(key, []):
                self.validate_params(params, Klass)
                device = self.make_device(params, Klass)
                try:
                    # Maybe its a list of devices?
                    devices.extend(device)
                except TypeError:
                    # No, assume it's just a single device then
                    devices.append(device)
        # Save devices for connecting to later
        self.devices.extend(devices)
        return devices

    def validate_params(self, params, Klass):
        """Check that parameters match a Device class's initializer."""
        sig = inspect.signature(Klass)
        has_kwargs = any(
            [param.kind == param.VAR_KEYWORD for param in sig.parameters.values()]
        )
        # Make sure we're not missing any required parameters
        for key, sig_param in sig.parameters.items():
            # Check for missing parameters
            param_missing = key not in params
            param_required = (
                sig_param.default is sig_param.empty
                and sig_param.kind != sig_param.VAR_KEYWORD
            )
            if param_missing and param_required:
                raise InvalidConfiguration(
                    f"Missing required key '{key}' for {Klass}: {params}"
                )
            # Check types
            if not param_missing:
                try:
                    correct_type = isinstance(params[key], sig_param.annotation)
                    has_type = not issubclass(sig_param.annotation, inspect._empty)
                except TypeError:
                    correct_type = False
                    has_type = False
                if has_type and not correct_type:
                    raise InvalidConfiguration(
                        f"Incorrect type for {Klass} key '{key}': "
                        f"expected `{sig_param.annotation}` but got "
                        f"`{type(params[key])}`."
                    )

    def make_device(self, params, Klass):
        """Create the devices from their parameters."""
        # Mock threaded ophyd devices if necessary
        try:
            is_threaded_device = issubclass(Klass, ThreadedDevice)
        except TypeError:
            is_threaded_device = False
        if is_threaded_device and not self.hardware_is_present:
            Klass = make_fake_device(Klass)
        # Turn the parameters into pure python objects
        kwargs = {}
        for key, param in params.items():
            if isinstance(param, tomlkit.items.Item):
                kwargs[key] = param.unwrap()
            else:
                kwargs[key] = param
        # Check if we need to injec the registry
        extra_params = {}
        sig = inspect.signature(Klass)
        if "registry" in sig.parameters.keys():
            kwargs["registry"] = self.registry
        # Create the device
        result = Klass(**kwargs)
        return result

    async def connect(
        self,
        mock: bool = False,
        timeout: float = DEFAULT_TIMEOUT,
        force_reconnect: bool = False,
        return_exceptions: bool = False,
    ):
        """Connect all Devices.

        Contains a timeout that gets propagated to device.connect methods.

        Parameters
        ----------
        mock:
          If True then use ``MockSignalBackend`` for all Signals
        timeout:
          Time to wait before failing with a TimeoutError.
        force_reconnect
          Force the signals to establish a new connection.
        """
        t0 = time.monotonic()
        # Sort out which devices are which
        threaded_devices = []
        async_devices = []
        for device in self.devices:
            if hasattr(device, "connect"):
                async_devices.append(device)
            else:
                threaded_devices.append(device)
        # Connect to async devices
        aws = (
            dev.connect(mock=mock, timeout=timeout, force_reconnect=force_reconnect)
            for dev in async_devices
        )
        results = await asyncio.gather(*aws, return_exceptions=True)
        # Filter out the disconnected devices
        new_devices = []
        exceptions = {}
        for device, result in zip(async_devices, results):
            if result is None:
                log.debug(f"Successfully connected device {device.name}")
                new_devices.append(device)
            else:
                # Unexpected exception, raise it so it can be handled
                log.debug(f"Failed connection for device {device.name}")
                exceptions[device.name] = result
        # Connect to threaded devices
        timeout_reached = False
        while not timeout_reached and len(threaded_devices) > 0:
            # Remove any connected devices for the running list
            connected_devices = [
                dev for dev in threaded_devices if getattr(dev, "connected", True)
            ]
            new_devices.extend(connected_devices)
            threaded_devices = [
                dev for dev in threaded_devices if dev not in connected_devices
            ]
            # Tick the clock for the next round through the while loop
            await asyncio.sleep(min((0.05, timeout / 10.0)))
            timeout_reached = (time.monotonic() - t0) > timeout
        # Add disconnected devices to the exception list
        for device in threaded_devices:
            try:
                device.wait_for_connection(timeout=0)
            except TimeoutError as exc:
                exceptions[device.name] = NotConnected(str(exc))
        # Raise exceptions if any were present
        if return_exceptions:
            return new_devices, exceptions
        if len(exceptions) > 0:
            raise NotConnected(exceptions)
        return new_devices

    async def load(
        self,
        connect: bool = True,
        device_classes: Mapping | None = None,
        config_files: Sequence[Path] | None = None,
        return_exceptions: bool = False,
    ):
        """Load instrument specified in config files.

        Unless, explicitly overridden by the *config_files* argument,
        configuration files are read from the environmental variable
        HAVEN_CONFIG_FILES (separated by ':').

        Parameters
        ==========
        connect
          If true, establish connections for the devices now.
        device_classes
          A temporary set of device classes to use for this call
          only. Overrides any device classes given during
          initalization.
        config_files
          I list of file paths that will be loaded. If omitted, those
          files listed in HAVEN_CONFIG_FILES will be used.
        return_exceptions
          If true, exceptions will be returned for further processing,
          otherwise, exceptions will be raised (default).

        """
        self.devices = []
        # Decide which config files to use
        if config_files is None:
            env_key = "HAVEN_CONFIG_FILES"
            if env_key in os.environ.keys():
                config_files = os.environ.get("HAVEN_CONFIG_FILES", "")
                config_files = [Path(fp) for fp in config_files.split(":")]
            else:
                config_files = [
                    Path(__file__).parent.resolve() / "iconfig_testing.toml"
                ]
        # Load the instrument from config files
        old_classes = self.device_classes
        try:
            # Temprary override of device classes
            if device_classes is not None:
                self.device_classes = device_classes
            # Parse TOML files
            for fp in config_files:
                with open(fp, mode="tr", encoding="utf-8") as fd:
                    self.parse_toml_file(fd)
        finally:
            self.device_classes = old_classes
        # Connect the devices
        if connect:
            new_devices, exceptions = await self.connect(
                mock=not self.hardware_is_present, return_exceptions=True
            )
        else:
            new_devices = self.devices
            exceptions = []
        # Registry devices
        for device in new_devices:
            self.registry.register(device)
        # Raise exceptions
        if return_exceptions:
            return exceptions
        elif len(exceptions) > 0:
            raise NotConnected(exceptions)


class HavenInstrument(Instrument):
    async def load(
        self,
        config: Mapping = None,
        wait_for_connection: bool = True,
        timeout: int = 5,
        return_devices: bool = False,
        reset_devices: bool = True,
    ):
        """Load the beamline instrumentation.

        This function will reach out and query various IOCs for motor
        information based on the information in *config* (see
        ``iconfig_default.toml`` for examples). Based on the
        configuration, it will create Ophyd devices and register them with
        *registry*.

        Parameters
        ==========
        registry:
          The registry into which the ophyd devices will be placed.
        config:
          The beamline configuration read in from TOML files. Mostly
          useful for testing.
        wait_for_connection
          If true, only connected devices will be kept.
        timeout
          How long to wait for if *wait_for_connection* is true.
        return_devices
          If true, return the newly loaded devices when complete.
        reset_registry
          If true, existing devices will be removed before loading.

        """
        if reset_devices:
            self.registry.clear()
        exceptions = await super().load(return_exceptions=True)
        # VME-style Motors happen later so duplicate motors can be
        # removed
        exceptions.update(
            await super().load(
                device_classes={"motors": load_motors}, return_exceptions=True
            )
        )
        # Handle connection exceptions
        if len(exceptions) > 0:
            raise NotConnected(exceptions)
        # Return the final list
        if return_devices:
            return instrument.devices
        else:
            return instrument


beamline = HavenInstrument(
    {
        # Ophyd-async devices
        "ion_chamber": IonChamber,
        "high_heat_load_mirror": HighHeatLoadMirror,
        "kb_mirrors": KBMirrors,
        "xy_stage": XYStage,
        "table": Table,
        "aerotech_stage": AerotechStage,
        "motor": Motor,
        "energy": EnergyPositioner,
        # Threaded ophyd devices
        "blade_slits": BladeSlits,
        "aperture_slits": ApertureSlits,
        "capillary_heater": CapillaryHeater,
        "power_supply": NHQ203MChannel,
        "synchrotron": ApsMachine,
        "robot": Robot,
        "pfcu4": PFCUFilterBank,  # <-- fails if mocked
        "pss_shutter": PssShutter,
        "xspress": make_xspress_device,
        "dxp": make_dxp_device,
        "beamline_manager": BeamlineManager,
        "area_detector": make_area_detector,
        "scaler": MultiChannelScaler,
    },
)



import pytest
from bluesky_queueserver_api import BPlan
from qtpy import QtCore

from firefly.plans.move_motor_window import MoveMotorDisplay


@pytest.fixture()
async def display(qtbot, sim_registry, sync_motors, async_motors):
    display = MoveMotorDisplay()
    qtbot.addWidget(display)
    await display.update_devices(sim_registry)
    return display


@pytest.mark.asyncio
async def test_move_motor_plan_queued(display, qtbot):
    display.ui.run_button.setEnabled(True)

    # uncheck relative
    display.ui.relative_scan_checkbox.setChecked(False)

    await display.update_regions(2)

    # set up a test motor 1
    display.regions[0].motor_box.combo_box.setCurrentText("async_motor_1")
    display.regions[0].position_line_edit.setText("111")

    # set up a test motor 2
    display.regions[1].motor_box.combo_box.setCurrentText("sync_motor_2")
    display.regions[1].position_line_edit.setText("222")

    expected_item = BPlan(
        "mv",
        "async_motor_1",
        111.0,
        "sync_motor_2",
        222.0,
    )

    def check_item(item):
        print(item.to_dict())
        return item.to_dict() == expected_item.to_dict()

    # Click the run button and see if the plan is queued
    with qtbot.waitSignal(
        display.queue_item_submitted, timeout=1000, check_params_cb=check_item
    ):
        qtbot.mouseClick(display.ui.run_button, QtCore.Qt.LeftButton)



"""An assortment of tests for plans defined in Haven.

More critical plans (e.g. xafs_scan) are tested in other test files.

"""

import warnings

import pytest
from ophyd import sim

from haven import align_slits


def test_align_slits(RE):
    """Check the plan to aligns the slits."""
    # Prepare a fake detector and slit motor
    slit_motor = sim.motor
    I0 = sim.SynGauss(
        "det",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"detectors"},
    )
    # Execute the plan
    RE(align_slits(slit_motors=[slit_motor], ion_chamber=I0))
    # Check that the slit positions have been set
    assert slit_motor.position == pytest.approx(-0.5)


def test_warn_poor_fit(RE):
    """Check that the plan emits a warning when no good fit is detected."""
    # Prepare a fake detector and slit motor
    slit_motor = sim.motor
    I0 = sim.SynGauss(
        "det",
        sim.motor,
        "motor",
        center=-0.5,
        noise="poisson",
        noise_multiplier=100,
        Imax=1,
        sigma=1,
        labels={"detectors"},
    )
    # Execute the plan, catching warnings
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        RE(align_slits(slit_motors=[slit_motor], ion_chamber=I0))
        # Check that a warning was raised
        assert len(w) >= 1
        messages = [str(w_.message) for w_ in w]
        target_message = "Poor fit while centering motor"
        assert any([target_message in msg for msg in messages])


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import argparse
import asyncio
import cProfile
import logging
import pstats
import sys
import time
from pathlib import Path

from pydm import config
from pydm.utilities import setup_renderer
from qasync import QEventLoop
from qtpy import QtCore
from qtpy.QtGui import QPixmap
from qtpy.QtWidgets import QApplication, QSplashScreen, QStyleFactory

from firefly.controller import FireflyController


def main(default_fullscreen=False, default_display="status"):
    log = logging.getLogger("")
    handler = logging.StreamHandler()
    formatter = logging.Formatter("[%(asctime)s] [%(levelname)-8s] - %(message)s")
    handler.setFormatter(formatter)
    log.addHandler(handler)
    log.setLevel("INFO")
    handler.setLevel("INFO")

    try:
        """
        We must import QtWebEngineWidgets before creating a QApplication
        otherwise we get the following error if someone adds a WebView at Designer:
        ImportError: QtWebEngineWidgets must be imported before a QCoreApplication instance is created
        """
    except ImportError:
        log.debug("QtWebEngine is not supported.")

    # Set up splash screen
    app = QApplication(sys.argv)
    im_dir = Path(__file__).parent.resolve()
    im_fp = str(im_dir / "splash.png")
    pixmap = QPixmap(im_fp)
    splash = QSplashScreen(pixmap, QtCore.Qt.WindowStaysOnTopHint)
    splash.show()
    app.processEvents()
    for i in range(10):
        time.sleep(0.01)
    app.processEvents()

    # Set EPICS as the default protocol
    config.DEFAULT_PROTOCOL = "ca"

    ui_folder = Path(__file__).parent.resolve()

    setup_renderer()

    parser = argparse.ArgumentParser(description="Python Display Manager")
    parser.add_argument(
        "display",
        help="The name of a Firefly display to open.",
        nargs="?",
        default=default_display,
    )
    parser.add_argument(
        "--no-instrument",
        action="store_true",
        help=(
            "Do not try to create devices. Useful for development if much beamline"
            " hardware is offline."
        ),
    )
    parser.add_argument(
        "--no-queue",
        action="store_true",
        help=(
            "Do not try to connect to the queueserver. Useful for development if queueserver"
            " hardware is offline."
        ),
    )
    parser.add_argument(
        "--perfmon",
        action="store_true",
        help="Enable performance monitoring," + " and print CPU usage to the terminal.",
    )
    parser.add_argument(
        "--profile",
        action="store_true",
        help="Enable cProfile function profiling, printing on exit.",
    )
    parser.add_argument(
        "--hide-menu-bar",
        action="store_true",
        help="Start PyDM with the menu bar hidden.",
    )
    parser.add_argument(
        "--hide-status-bar",
        action="store_true",
        help="Start PyDM with the status bar hidden.",
    )
    if default_fullscreen:
        parser.add_argument(
            "--no-fullscreen",
            dest="fullscreen",
            action="store_false",
            help="Start Firefly in normal (non-fullscreen) mode.",
        )
    else:
        parser.add_argument(
            "--fullscreen",
            action="store_true",
            help="Start Firefly in full screen mode.",
        )
    parser.add_argument(
        "--read-only", action="store_true", help="Start Firefly in a Read-Only mode."
    )
    parser.add_argument(
        "--log_level",
        help="Configure level of log display",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="INFO",
    )
    # parser.add_argument(
    #     "--version",
    #     action="version",
    #     version="Firefly {version}".format(version=haven.__version__),
    #     help="Show Firefly's version number and exit.",
    # )
    parser.add_argument(
        "-m",
        "--macro",
        help="Specify macro replacements to use, in JSON object format."
        + "    Reminder: JSON requires double quotes for strings, "
        + "so you should wrap this whole argument in single quotes."
        + '  Example: -m \'{"sector": "LI25", "facility": "LCLS"}\''
        + "--or-- specify macro replacements as KEY=value pairs "
        + " using a comma as delimiter  If you want to uses spaces "
        + " after the delimiters or around the = signs, "
        + " wrap the entire set with quotes "
        + '  Example: -m "sector = LI25, facility=LCLS"',
    )
    parser.add_argument(
        "--stylesheet",
        help="Specify the full path to a CSS stylesheet file, which"
        + " can be used to customize the appearance of Firefly and"
        + " Qt widgets.",
        default=None,
    )
    parser.add_argument(
        "display_args",
        help="Arguments to be passed to the PyDM client application"
        + " (which is a QApplication subclass).",
        nargs=argparse.REMAINDER,
        default=None,
    )

    pydm_args = parser.parse_args()
    if pydm_args.profile:
        profile = cProfile.Profile()
        profile.enable()

    if pydm_args.log_level:
        log.setLevel(pydm_args.log_level)
        handler.setLevel(pydm_args.log_level)

    controller = FireflyController(
        display=pydm_args.display,
    )
    qss_file = Path(__file__).parent / "firefly.qss"
    app.setStyleSheet(qss_file.read_text())
    # apply_stylesheet(app.stylesheet_path, widget=main_window)
    log.info(f"Available styles: {QStyleFactory.keys()}")

    # Make it asynchronous
    event_loop = QEventLoop(app)
    asyncio.set_event_loop(event_loop)

    async def run_app():
        # Keep an event to know when the app has closed
        app_close_event = asyncio.Event()
        app.aboutToQuit.connect(app_close_event.set)
        # Define devices on the beamline (slow!)
        await controller.setup_instrument(load_instrument=not pydm_args.no_instrument)
        # Start any async clients that the controller needs
        if not pydm_args.no_queue:
            controller.start()
        # Get rid of the splash screen and show the first window
        splash.close()
        controller.show_default_window()
        # Wait on the application to finish
        await app_close_event.wait()

    # Start the actual asyncio event loop
    event_loop.run_until_complete(run_app())
    event_loop.close()

    if pydm_args.profile:
        profile.disable()
        stats = pstats.Stats(
            profile,
            stream=sys.stdout,
        ).sort_stats(pstats.SortKey.CUMULATIVE)
        stats.print_stats()


def cameras():
    main(default_fullscreen=True, default_display="cameras")


if __name__ == "__main__":
    main()


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest
from qtpy.QtCore import Qt

from firefly.detector_list import DetectorListView


@pytest.fixture()
async def view(qtbot, sim_registry, dxp):
    view = DetectorListView()
    qtbot.addWidget(view)
    await view.update_devices(sim_registry)
    return view


def test_detector_model(view):
    assert hasattr(view, "detector_model")
    assert view.detector_model.item(0).text() == "vortex_me4"


def test_selected_detectors(qtbot, view):
    """Do we get the list of detectors after they have been selected?"""
    # No detectors selected, so empty list
    assert view.selected_detectors() == []
    # Select a detector and see if the selection updates
    item = view.detector_model.item(0)
    assert item is not None
    rect = view.visualRect(item.index())
    with qtbot.waitSignal(view.selectionModel().selectionChanged):
        qtbot.mouseClick(view.viewport(), Qt.LeftButton, pos=rect.center())
    assert view.selected_detectors() == ["vortex_me4"]


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



"""Tools for modifying plans and data streams as they are generated."""

import getpass
import logging
import os
import socket
import warnings
from collections import ChainMap
from typing import Sequence, Union  # , Iterable

import epics
import pkg_resources
from bluesky.preprocessors import baseline_wrapper as bluesky_baseline_wrapper
from bluesky.preprocessors import finalize_wrapper, msg_mutator

# from bluesky.suspenders import SuspendBoolLow
from bluesky.utils import Msg, make_decorator

from . import __version__ as haven_version
from ._iconfig import load_config
from .exceptions import ComponentNotFound
from .instrument import beamline

log = logging.getLogger()


def baseline_wrapper(
    plan,
    devices: Union[Sequence, str] = [
        "motors",
        "power_supplies",
        "xray_sources",
        "APS",
        "baseline",
    ],
    name: str = "baseline",
):
    bluesky_baseline_wrapper.__doc__
    # Resolve devices
    devices = beamline.registry.findall(devices, allow_none=True)
    yield from bluesky_baseline_wrapper(plan=plan, devices=devices, name=name)


def get_version(pkg_name):
    return pkg_resources.get_distribution(pkg_name).version


VERSIONS = dict(
    apstools=get_version("apstools"),
    bluesky=get_version("bluesky"),
    databroker=get_version("databroker"),
    epics_ca=epics.__version__,
    epics=epics.__version__,
    haven=haven_version,
    h5py=get_version("h5py"),
    matplotlib=get_version("matplotlib"),
    numpy=get_version("numpy"),
    ophyd=get_version("ophyd"),
    pymongo=get_version("pymongo"),
)


def inject_haven_md_wrapper(plan):
    """Inject additional metadata into a run.

    This takes precedences over the original metadata dict in the event of
    overlapping keys, but it does not mutate the original metadata dict.
    (It uses ChainMap.)

    Parameters
    ----------
    plan : iterable or iterator
        a generator, list, or similar containing `Msg` objects

    """

    def _inject_md(msg):
        if msg.command != "open_run":
            return msg
        # Prepare the metadata dictionary
        config = load_config()
        md = {
            # Software versions
            "versions": VERSIONS,
            # Controls
            "EPICS_HOST_ARCH": os.environ.get("EPICS_HOST_ARCH"),
            "epics_libca": os.environ.get("PYEPICS_LIBCA"),
            "EPICS_CA_MAX_ARRAY_BYTES": os.environ.get("EPICS_CA_MAX_ARRAY_BYTES"),
            # Facility
            "beamline_id": config["beamline"]["name"],
            "facility_id": ", ".join(cfg["name"] for cfg in config["synchrotron"]),
            "xray_source": config["xray_source"]["type"],
            # Computer
            "login_id": f"{getpass.getuser()}@{socket.gethostname()}",
            "pid": os.getpid(),
            # User supplied
            "sample_name": "",
            # Bluesky
            "parameters": "",
            "purpose": "",
        }
        # Get metadata from the beamline scheduling system (bss)
        try:
            bss = beamline.registry.find(name="bss")
        except ComponentNotFound:
            if config["beamline"]["hardware_is_present"]:
                wmsg = "Could not find bss device, metadata may be missing."
                warnings.warn(wmsg)
                log.warning(wmsg)
            bss_md = None
        else:
            bss_md = bss.get()
            md.update(
                {
                    "proposal_id": bss_md.proposal.proposal_id,
                    "proposal_title": bss_md.proposal.title,
                    "proposal_users": bss_md.proposal.user_last_names,
                    "proposal_user_badges": bss_md.proposal.user_badges,
                    "esaf_id": bss_md.esaf.esaf_id,
                    "esaf_title": bss_md.esaf.title,
                    "esaf_users": bss_md.esaf.user_last_names,
                    "esaf_user_badges": bss_md.esaf.user_badges,
                    "mail_in_flag": bss_md.proposal.mail_in_flag,
                    "proprietary_flag": bss_md.proposal.proprietary_flag,
                    "bss_aps_cycle": bss_md.esaf.aps_cycle,
                    "bss_beamline_name": bss_md.proposal.beamline_name,
                }
            )
        # Update the message
        msg = msg._replace(kwargs=ChainMap(msg.kwargs, md))
        return msg

    return (yield from msg_mutator(plan, _inject_md))


def shutter_suspend_wrapper(plan, shutter_signals=None):
    """
    Install suspenders to the RunEngine, and remove them at the end.

    Parameters
    ----------
    plan : iterable or iterator
        a generator, list, or similar containing `Msg` objects
    suspenders : suspender or list of suspenders
        Suspenders to use for the duration of the wrapper

    Yields
    ------
    msg : Msg
        messages from plan, with 'install_suspender' and 'remove_suspender'
        messages inserted and appended
    """
    if shutter_signals is None:
        shutters = beamline.registry.findall("shutters", allow_none=True)
        shutter_signals = [s.pss_state for s in shutters]
    # Create a suspender for each shutter
    suspenders = []

    ###################################################
    # Temporarily disabled for technical commissioning
    ###################################################
    # for sig in shutter_signals:
    #     suspender = SuspendBoolLow(sig, sleep=3.0)
    #     suspenders.append(suspender)
    # if not isinstance(suspenders, Iterable):
    #     suspenders = [suspenders]

    def _install():
        for susp in suspenders:
            yield Msg("install_suspender", None, susp)

    def _remove():
        for susp in suspenders:
            yield Msg("remove_suspender", None, susp)

    def _inner_plan():
        yield from _install()
        return (yield from plan)

    return (yield from finalize_wrapper(_inner_plan(), _remove()))


baseline_decorator = make_decorator(baseline_wrapper)
shutter_suspend_decorator = make_decorator(shutter_suspend_wrapper)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



# import os
# import warnings
# from pathlib import Path
# from unittest.mock import MagicMock
# from types import SimpleNamespace
# from collections import OrderedDict
# from subprocess import Popen, TimeoutExpired, PIPE
# import subprocess
# import shutil
# import time
# from unittest import mock
# import asyncio

# import pytest

# import ophyd
# from ophyd import DynamicDeviceComponent as DDC, Kind
# from ophyd.sim import (
#     instantiate_fake_device,
#     make_fake_device,
#     fake_device_cache,
#     FakeEpicsSignal,
# )
# from pydm.data_plugins import add_plugin

# import haven
# from haven.simulated_ioc import simulated_ioc
# from haven import load_config, registry
# from haven._iconfig import beamline_connected as _beamline_connected
# from haven.instrument.aerotech import AerotechFlyer, AerotechStage
# from haven.instrument.aps import ApsMachine
# from haven.instrument.shutter import Shutter
# from haven.instrument.camera import AravisDetector
# from haven.instrument.delay import EpicsSignalWithIO
# from haven.instrument.dxp import DxpDetector, add_mcas as add_dxp_mcas
# from haven.instrument.ion_chamber import IonChamber
# from haven.instrument.xspress import Xspress3Detector, add_mcas as add_xspress_mcas
# from firefly.application import FireflyApplication
# from firefly.ophyd_plugin import OphydPlugin
# from firefly.main_window import FireflyMainWindow


# # IOC_SCOPE = "function"
# IOC_SCOPE = "session"


# @pytest.fixture(scope=IOC_SCOPE)
# def ioc_undulator(request):
#     prefix = "ID255:"
#     pvs = dict(energy=f"{prefix}Energy.VAL")
#     return run_fake_ioc(
#         module_name="haven.tests.ioc_undulator",
#         name="Fake undulator IOC",
#         prefix=prefix,
#         pvs=pvs,
#         pv_to_check=pvs["energy"],
#         request=request,
#     )


# @pytest.fixture(scope=IOC_SCOPE)
# def ioc_camera(request):
#     prefix = "255idgigeA:"
#     pvs = dict(
#         cam_acquire=f"{prefix}cam1:Acquire",
#         cam_acquire_busy=f"{prefix}cam1:AcquireBusy",
#     )
#     return run_fake_ioc(
#         module_name="haven.tests.ioc_area_detector",
#         name="Fake IOC for a simulated machine vision camera",
#         prefix=prefix,
#         pvs=pvs,
#         pv_to_check=pvs["cam_acquire_busy"],
#         request=request,
#     )


# @pytest.fixture(scope=IOC_SCOPE)
# def ioc_area_detector(request):
#     prefix = "255idSimDet:"
#     pvs = dict(
#         cam_acquire=f"{prefix}cam1:Acquire",
#         cam_acquire_busy=f"{prefix}cam1:AcquireBusy",
#     )
#     return run_fake_ioc(
#         module_name="haven.tests.ioc_area_detector",
#         name="Fake IOC for a simulated area detector",
#         prefix=prefix,
#         pvs=pvs,
#         pv_to_check=pvs["cam_acquire_busy"],
#         request=request,
#     )


# @pytest.fixture(scope=IOC_SCOPE)
# def ioc_bss(request):
#     prefix = "255idc:bss:"
#     pvs = dict(
#         esaf_id=f"{prefix}esaf:id",
#         esaf_cycle=f"{prefix}esaf:cycle",
#         proposal_id=f"{prefix}proposal:id",
#     )
#     return run_fake_ioc(
#         module_name="haven.tests.ioc_apsbss",
#         name="Fake IOC for APS beamline scheduling system (BSS)",
#         prefix=prefix,
#         pvs=pvs,
#         pv_to_check=pvs["esaf_cycle"],
#         request=request,
#     )


# def run_fake_ioc(
#     module_name,
#     prefix: str,
#     request,
#     name="Fake IOC",
#     pvs=None,
#     pv_to_check: str = None,
# ):
#     if pvs is None:
#         pvs = {}
#     pytest.importorskip("caproto.tests.conftest")
#     from caproto.tests.conftest import run_example_ioc, poll_readiness

#     process = run_example_ioc(
#         module_name=module_name,
#         request=request,
#         pv_to_check=None,
#         args=("--prefix", prefix, "--list-pvs", "-v"),
#         very_verbose=False,
#     )
#     # Verify the IOC started
#     if pv_to_check is not None:
#         poll_timeout, poll_attempts = 1.0, 30
#         poll_readiness(
#             pv_to_check, timeout=poll_timeout, attempts=poll_attempts, process=process
#         )
#     return SimpleNamespace(
#         process=process, prefix=prefix, name=name, pvs=pvs, type="caproto"
#     )


# @pytest.fixture(scope=IOC_SCOPE)
# def ioc_scaler(request):
#     prefix = "255idVME:scaler1"
#     pvs = dict(calc=f"{prefix}_calc2.VAL")
#     return run_fake_ioc(
#         module_name="haven.tests.ioc_scaler",
#         name="Fake scaler IOC",
#         prefix=prefix,
#         pvs=pvs,
#         pv_to_check=pvs["calc"],
#         request=request,
#     )


# @pytest.fixture(scope=IOC_SCOPE)
# def ioc_ptc10(request):
#     prefix = "255idptc10:"
#     pvs = dict(
#         pid1_voltage=f"{prefix}5A:output",
#         pid1_voltage_rbv=f"{prefix}5A:output_RBV",
#         tc1_temperature=f"{prefix}2A:temperature",
#     )
#     return run_fake_ioc(
#         module_name="haven.tests.ioc_ptc10",
#         name="Fake PTC10 temperature controller IOC",
#         prefix=prefix,
#         pvs=pvs,
#         pv_to_check=pvs["tc1_temperature"],
#         request=request,
#     )


# @pytest.fixture(scope="session")
# def pydm_ophyd_plugin():
#     return add_plugin(OphydPlugin)


# @pytest.fixture(scope=IOC_SCOPE)
# def ioc_motor(request):
#     prefix = "255idVME:"
#     pvs = dict(m1=f"{prefix}m1", m2=f"{prefix}m2", m3=f"{prefix}m3", m4=f"{prefix}m4")
#     return run_fake_ioc(
#         module_name="haven.tests.ioc_motor",
#         name="Fake motor IOC",
#         prefix=prefix,
#         pvs=pvs,
#         pv_to_check=pvs["m1"],
#         request=request,
#     )


# @pytest.fixture(scope=IOC_SCOPE)
# def ioc_preamp(request):
#     prefix = "255idc:"
#     pvs = dict(
#         preamp1_sens_num=f"{prefix}SR01:sens_num",
#         preamp2_sens_num=f"{prefix}SR02:sens_num",
#         preamp3_sens_num=f"{prefix}SR03:sens_num",
#         preamp4_sens_num=f"{prefix}SR04:sens_num",
#         preamp1_sens_unit=f"{prefix}SR01:sens_unit",
#         preamp2_sens_unit=f"{prefix}SR02:sens_unit",
#         preamp3_sens_unit=f"{prefix}SR03:sens_unit",
#         preamp4_sens_unit=f"{prefix}SR04:sens_unit",
#         preamp1_offset_num=f"{prefix}SR01:offset_num",
#         preamp2_offset_num=f"{prefix}SR02:offset_num",
#         preamp3_offset_num=f"{prefix}SR03:offset_num",
#         preamp4_offset_num=f"{prefix}SR04:offset_num",
#         preamp1_offset_unit=f"{prefix}SR01:offset_unit",
#         preamp2_offset_unit=f"{prefix}SR02:offset_unit",
#         preamp3_offset_unit=f"{prefix}SR03:offset_unit",
#         preamp4_offset_unit=f"{prefix}SR04:offset_unit",
#     )
#     return run_fake_ioc(
#         module_name="haven.tests.ioc_preamp",
#         name="Fake preamp IOC",
#         prefix=prefix,
#         pvs=pvs,
#         pv_to_check=pvs["preamp1_sens_num"],
#         request=request,
#     )


# @pytest.fixture(scope=IOC_SCOPE)
# def ioc_simple(request):
#     prefix = "simple:"
#     pvs = dict(
#         A=f"{prefix}A",
#         B=f"{prefix}B",
#         C=f"{prefix}C",
#     )
#     pv_to_check = pvs["A"]
#     return run_fake_ioc(
#         module_name="haven.tests.ioc_simple",
#         name="Fake simple IOC",
#         prefix=prefix,
#         pvs=pvs,
#         pv_to_check=pv_to_check,
#         request=request,
#     )


# @pytest.fixture(scope=IOC_SCOPE)
# def ioc_mono(request):
#     prefix = "255idMono:"
#     pvs = dict(
#         bragg=f"{prefix}ACS:m3",
#         energy=f"{prefix}Energy",
#         id_tracking=f"{prefix}ID_tracking",
#         id_offset=f"{prefix}ID_offset",
#     )
#     return run_fake_ioc(
#         module_name="haven.tests.ioc_mono",
#         name="Fake mono IOC",
#         prefix=prefix,
#         pvs=pvs,
#         pv_to_check=pvs["energy"],
#         request=request,
#     )


# @pytest.fixture(scope=IOC_SCOPE)
# def ioc_dxp(request):
#     prefix = "255idDXP:"
#     pvs = dict(acquiring=f"{prefix}Acquiring")
#     return run_fake_ioc(
#         module_name="haven.tests.ioc_dxp",
#         name="Fake DXP-based detector IOC",
#         prefix=prefix,
#         pvs=pvs,
#         pv_to_check=pvs["acquiring"],
#         request=request,
#     )


# # Simulated devices
# @pytest.fixture()
# def queue_app(ffapp):
#     """An application that is set up to interact (fakely) with the queue
#     server.

#     """
#     warnings.warn("queue_app is deprecated, just use ffapp instead.")
#     return ffapp


# @pytest.fixture()
# def sim_vortex(dxp):
#     warnings.warn("sim_vortex is deprecated, just use ``dxp`` instead.")
#     return dxp



import pytest

from haven.devices.xia_pfcu import PFCUFilter, PFCUFilterBank, PFCUShutter, ShutterState


@pytest.fixture()
def shutter(xia_shutter):
    yield xia_shutter


@pytest.fixture()
def shutter_bank(xia_shutter_bank):
    yield xia_shutter_bank


def test_shutter_factory():
    """Check that a shutter device is created if requested."""
    filterbank = PFCUFilterBank(
        prefix="255id:pfcu0:", name="filter_bank_0", shutters=[(2, 3)]
    )  #
    assert filterbank.prefix == "255id:pfcu0:"
    assert filterbank.name == "filter_bank_0"
    assert hasattr(filterbank, "shutters")
    assert isinstance(filterbank.shutters.shutter_0, PFCUShutter)
    assert hasattr(filterbank, "shutters")
    assert isinstance(filterbank.filters.filter1, PFCUFilter)
    assert filterbank.filters.filter1.prefix == "255id:pfcu0:filter1"
    assert isinstance(filterbank.filters.filter4, PFCUFilter)
    assert not hasattr(filterbank.filters, "filter2")
    assert not hasattr(filterbank.filters, "filter3")


def test_pfcu_shutter_signals(shutter):
    # Check initial state
    assert shutter.top_filter.setpoint.get() == 0
    assert shutter.bottom_filter.setpoint.get() == 0


def test_pfcu_shutter_readback(shutter):
    # Set the shutter position
    readback = shutter.parent.parent.readback
    readback._readback = 0b0010
    readback._run_subs(sub_type=readback._default_sub)
    # Check that the readback signal gets updated
    assert shutter.readback.get() == ShutterState.OPEN


def test_pfcu_shutter_bank_mask(shutter_bank):
    """A bit-mask used for determining how to set the filter bank."""
    shutter = shutter_bank.shutters.shutter_0
    assert shutter.setpoint.top_mask() == 0b0001
    assert shutter.setpoint.bottom_mask() == 0b0010


def test_pfcu_shutter_open(shutter_bank):
    """If the PFCU filter bank is available, open both blades simultaneously."""
    shutter = shutter_bank.shutters.shutter_0
    # Set the other filters on the filter bank
    shutter_bank.readback._readback = 0b0100
    # Open the shutter, and check that the filterbank was set
    shutter.setpoint.set(ShutterState.OPEN).wait(timeout=1)
    assert shutter_bank.setpoint.get() == 0b0110


def test_pfcu_shutter_close(shutter_bank):
    """If the PFCU filter bank is available, open both blades simultaneously."""
    shutter = shutter_bank.shutters.shutter_0
    # Set the other filters on the filter bank
    shutter_bank.readback._readback = 0b0100
    # Open the shutter, and check that the filterbank was set
    shutter.setpoint.set(ShutterState.CLOSED).wait(timeout=1)
    assert shutter_bank.setpoint.get() == 0b0101


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from unittest import mock

import numpy as np
import pydm
import pyqtgraph
import pytest

from firefly.area_detector_viewer import AreaDetectorViewerDisplay


@pytest.fixture()
def display(qtbot, sim_camera):
    display = AreaDetectorViewerDisplay(macros={"AD": sim_camera.name})
    qtbot.addWidget(display)
    return display


def test_image_plotting(display):
    assert isinstance(display.image_view, pyqtgraph.ImageView)
    assert isinstance(display.image_channel, pydm.PyDMChannel)
    # Give it some grayscale data
    source_img = np.random.randint(0, 256, size=(54, 64))
    # For some reason, PyDMConnection mis-shapes the data
    reshape_img = np.reshape(source_img, source_img.shape[::-1])
    display.image_channel.value_slot(reshape_img)
    # See if we get a plottable image out
    new_image = display.image_view.getImageItem()
    assert new_image.image is not None
    assert new_image.image.shape == (54, 64)
    np.testing.assert_equal(new_image.image, source_img)
    # Give it some RGB data
    source_img = np.random.randint(0, 256, size=(54, 64, 3))
    # For some reason, PyDMConnection mis-shapes the data
    reshape_img = np.reshape(source_img, source_img.shape[::-1])
    display.image_channel.value_slot(reshape_img)
    # See if we get a plottable image out
    new_image = display.image_view.getImageItem()
    assert new_image.image.shape == (54, 64, 3)
    display.image_channel.disconnect()


def test_caqtdm_window(display, sim_camera):
    display._open_caqtdm_subprocess = mock.MagicMock()
    # Launch the caqtdm display
    display.launch_caqtdm()
    display._open_caqtdm_subprocess.assert_called_once_with(
        f"start_{sim_camera.prefix}_caqtdm"
    )


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------
